from tornado.web import RedirectHandler as TornadoRedirectHandler
from tornado.web import RequestHandler


class EnhanceHeadersMixin:
    """
    EnhanceHeadersMixin allows to add security headers to GUI urls.
    """

    def set_header_strict_transport_security(self) -> None:
        # rhbz#1558063 rhbz#2097392
        # The HTTP Strict-Transport-Security response header (often abbreviated
        # as HSTS) lets a web site tell browsers that it should only be
        # accessed using HTTPS, instead of using HTTP.
        # Do not set "includeSubDomains" as that would affect all web sites and
        # applications running on any subdomain of the domain where pcs web UI
        # is running. The fact that pcs web UI runs on a specific port doesn't
        # matter, subdomains would still be affected.
        self.set_header("Strict-Transport-Security", "max-age=63072000")

    def set_header_nosniff_content_type(self) -> None:
        # The X-Content-Type-Options response HTTP header is a marker used by
        # the server to indicate that the MIME types advertised in the
        # Content-Type headers should not be changed and be followed. This
        # allows to opt-out of MIME type sniffing, or, in other words, it is a
        # way to say that the webmasters knew what they were doing.
        self.set_header("X-Content-Type-Options", "nosniff")

    def clear_header_server(self) -> None:
        # The Server header describes the software used by the origin server
        # that handled the request — that is, the server that generated the
        # response.
        #
        # rhbz 2058278
        # When a HTTP request is made against a cluster node running pcsd, the
        # HTTP response contains HTTP Server name in its headers.
        # This is perceived as a security threat.
        self.clear_header("Server")

    def set_header_frame_options(self) -> None:
        # The X-Frame-Options HTTP response header can be used to indicate
        # whether or not a browser should be allowed to render a page in a
        # <frame>, <iframe> or <object> . Sites can use this to avoid
        # clickjacking attacks, by ensuring that their content is not embedded
        # into other sites.
        self.set_header("X-Frame-Options", "SAMEORIGIN")

    def set_header_content_security_policy(self) -> None:
        # The HTTP Content-Security-Policy (CSP) frame-ancestors directive
        # specifies valid parents that may embed a page using <frame>,
        # <iframe>, <object>, <embed>, or <applet>.

        # Requested upstream:
        # https://lists.clusterlabs.org/pipermail/users/2020-May/027199.html

        # For now, I'm just setting this to the same value as already present
        # X-Frame-Options header to keep them consistent.
        self.set_header("Content-Security-Policy", "frame-ancestors 'self'")

    def set_header_xss_protection(self) -> None:
        # The HTTP X-XSS-Protection response header is a feature of Internet
        # Explorer, Chrome and Safari that stops pages from loading when they
        # detect reflected cross-site scripting (XSS) attacks. Although these
        # protections are largely unnecessary in modern browsers when sites
        # implement a strong Content-Security-Policy that disables the use of
        # inline JavaScript ('unsafe-inline'), they can still provide
        # protections for users of older web browsers that don't yet support
        # CSP.
        self.set_header("X-Xss-Protection", "1; mode=block")

    def set_header_cache_control(self) -> None:
        # rhbz#2097383
        # The point is to prevent stealing sensitive data (such as fence
        # devices passwords) from browser caches.

        # no-store - Any caches of any kind should not store this response.
        # no-cache - Response can be stored in caches, but the response must be
        #   validated with the origin server before each reuse. This was
        #   requested probably as a fallback for caches which don't support
        #   no-store.
        self.set_header("Cache-Control", "no-store, no-cache")
        self.set_header("Pragma", "no-cache")

    def clear_header_cache_control(self) -> None:
        # Revert headers to default to allow caching, useful e.g. for static
        # content.
        self.clear_header("Cache-Control")
        self.clear_header("Pragma")

    def set_header_referrer(self) -> None:
        # rhbz#2097391
        # Do not expose cluster configuration bits to other websites (URL may
        # contain name of nodes/resources/etc.). This is mostly future-proof
        # hardening as there are currently little to no links to external
        # sites.
        self.set_header("Referrer-Policy", "no-referrer")

    def set_default_headers(self) -> None:
        """
        Modifies automatic tornado headers for all responses (i.e. including
        errors).

        Method `initialize` is the place for setting headers only for success
        responses.
        """
        self.clear_header_server()
        self.set_header_strict_transport_security()
        self.set_header_nosniff_content_type()
        self.set_header_frame_options()
        self.set_header_content_security_policy()
        self.set_header_xss_protection()
        self.set_header_cache_control()
        self.set_header_referrer()


class BaseHandler(EnhanceHeadersMixin, RequestHandler):
    """
    BaseHandler modifies HTTP headers.
    """

    def data_received(self, chunk: bytes) -> None:
        # abstract method `data_received` does need to be overridden. This
        # method should be implemented to handle streamed request data.
        # BUT we currently do not plan to use it SO:
        pass


# abstract method `data_received` is not used in redirect:
# pylint: disable=abstract-method
class RedirectHandler(EnhanceHeadersMixin, TornadoRedirectHandler):
    """
    RedirectHandler with modified HTTP headers.
    """



from functools import partial
from unittest import (
    TestCase,
    mock,
)

from lxml import etree

from pcs.common.reports import codes as report_codes
from pcs.common.tools import Version
from pcs.lib.env import LibraryEnvironment

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_xml_equal
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.custom_mock import (
    TmpFileCall,
    TmpFileMock,
)
from pcs_test.tools.misc import create_setup_patch_mixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.xml import etree_to_str


def mock_tmpfile(filename):
    mock_file = mock.MagicMock()
    mock_file.name = rc(filename)
    return mock_file


SetupPatchMixin = create_setup_patch_mixin(
    partial(mock.patch.object, LibraryEnvironment)
)


class ManageCibAssertionMixin:
    def assert_raises_cib_error(self, callable_obj, message):
        with self.assertRaises(AssertionError) as context_manager:
            callable_obj()
            self.assertEqual(str(context_manager.exception), message)

    def assert_raises_cib_not_loaded(self, callable_obj):
        self.assert_raises_cib_error(callable_obj, "CIB has not been loaded")

    def assert_raises_cib_already_loaded(self, callable_obj):
        self.assert_raises_cib_error(
            callable_obj, "CIB has already been loaded"
        )

    def assert_raises_cib_loaded_cannot_custom(self, callable_obj):
        self.assert_raises_cib_error(
            callable_obj, "CIB has been loaded, cannot push custom CIB"
        )


class IsCibLive(TestCase):
    def test_is_live_when_no_cib_data_specified(self):
        env_assist, _ = get_env_tools(test_case=self)
        self.assertTrue(env_assist.get_env().is_cib_live)

    def test_is_not_live_when_cib_data_specified(self):
        env_assist, config = get_env_tools(test_case=self)
        config.env.set_cib_data("<cib/>")
        self.assertFalse(env_assist.get_env().is_cib_live)


class UpgradeCib(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_get_and_push_cib_version_upgrade_needed(self):
        (
            self.config.runner.cib.load(
                name="load_cib_old", filename="cib-empty-3.1.xml"
            )
            .runner.cib.upgrade()
            .runner.cib.load(filename="cib-empty-3.2.xml")
        )
        env = self.env_assist.get_env()
        env.get_cib(Version(3, 2, 0))

        self.env_assist.assert_reports(
            [fixture.info(report_codes.CIB_UPGRADE_SUCCESSFUL)]
        )

    def test_get_and_push_cib_version_upgrade_not_needed(self):
        self.config.runner.cib.load(filename="cib-empty-3.2.xml")
        env = self.env_assist.get_env()
        env.get_cib(Version(3, 1, 0))

    def test_nice_to_have_lower_than_required(self):
        (
            self.config.runner.cib.load(
                name="load_cib_old", filename="cib-empty-3.1.xml"
            )
            .runner.cib.upgrade()
            .runner.cib.load(filename="cib-empty-3.3.xml")
        )
        env = self.env_assist.get_env()
        env.get_cib(Version(3, 3, 0), Version(3, 2, 0))

        self.env_assist.assert_reports(
            [fixture.info(report_codes.CIB_UPGRADE_SUCCESSFUL)]
        )

    def test_nice_to_have_equal_required(self):
        (
            self.config.runner.cib.load(
                name="load_cib_old", filename="cib-empty-3.1.xml"
            )
            .runner.cib.upgrade()
            .runner.cib.load(filename="cib-empty-3.3.xml")
        )
        env = self.env_assist.get_env()
        env.get_cib(Version(3, 3, 0), Version(3, 3, 0))

        self.env_assist.assert_reports(
            [fixture.info(report_codes.CIB_UPGRADE_SUCCESSFUL)]
        )

    def test_satisfied_nice_to_have_greater_than_required(self):
        (
            self.config.runner.cib.load(
                name="load_cib_old", filename="cib-empty-3.1.xml"
            )
            .runner.cib.upgrade()
            .runner.cib.load(filename="cib-empty-3.3.xml")
        )
        env = self.env_assist.get_env()
        env.get_cib(Version(3, 2, 0), Version(3, 3, 0))

        self.env_assist.assert_reports(
            [fixture.info(report_codes.CIB_UPGRADE_SUCCESSFUL)]
        )

    def test_not_satisfied_nice_to_have_greater_than_required(self):
        (
            self.config.runner.cib.load(
                name="load_cib_old", filename="cib-empty-3.1.xml"
            )
            .runner.cib.upgrade()
            .runner.cib.load(filename="cib-empty-3.3.xml")
        )
        env = self.env_assist.get_env()
        env.get_cib(Version(3, 3, 0), Version(3, 4, 0))

        self.env_assist.assert_reports(
            [fixture.info(report_codes.CIB_UPGRADE_SUCCESSFUL)]
        )


class GetCib(TestCase, ManageCibAssertionMixin):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.tmp_file = "/fake/tmp/file"

    def test_raise_library_error_when_cibadmin_failed(self):
        stderr = "cibadmin: Connection to local file failed..."
        (
            self.config
            # Value of cib_data is unimportant here. This content is only put
            # into tempfile when the runner is not mocked. And content is then
            # loaded from tempfile by `cibadmin --local --query`. Runner is
            # mocked in tests so the value of cib_data is not in the fact used.
            .env.set_cib_data(
                "whatever", cib_tempfile=self.tmp_file
            ).runner.cib.load(
                returncode=203, stderr=stderr, env=dict(CIB_file=self.tmp_file)
            )
        )

        self.env_assist.assert_raise_library_error(
            self.env_assist.get_env().get_cib,
            [fixture.error(report_codes.CIB_LOAD_ERROR, reason=stderr)],
            expected_in_processor=False,
        )

    def test_returns_cib_from_cib_data(self):
        cib_filename = "cib-empty.xml"
        (
            self.config
            # Value of cib_data is unimportant here. See details in sibling test
            .env.set_cib_data(
                "whatever", cib_tempfile=self.tmp_file
            ).runner.cib.load(
                filename=cib_filename, env=dict(CIB_file=self.tmp_file)
            )
        )
        with open(rc(cib_filename)) as cib_file:
            assert_xml_equal(
                etree_to_str(self.env_assist.get_env().get_cib()),
                cib_file.read(),
            )

    def test_get_and_property(self):
        self.config.runner.cib.load()
        env = self.env_assist.get_env()
        self.assertEqual(env.get_cib(), env.cib)

    def test_property_without_get(self):
        env = self.env_assist.get_env()
        # need to use lambda because env.cib is a property
        self.assert_raises_cib_not_loaded(lambda: env.cib)

    def test_double_get(self):
        self.config.runner.cib.load()
        env = self.env_assist.get_env()
        env.get_cib()
        self.assert_raises_cib_already_loaded(env.get_cib)


class PushLoadedCib(TestCase, ManageCibAssertionMixin):
    # pylint: disable=too-many-public-methods
    wait_timeout = 10

    def setUp(self):
        self.tmpfile_old = "old.cib"
        self.tmpfile_new = "new.cib"
        self.load_cib_name = "load_cib"
        self.tmp_file_mock_obj = TmpFileMock(
            file_content_checker=assert_xml_equal,
        )
        self.addCleanup(self.tmp_file_mock_obj.assert_all_done)
        tmp_file_patcher = mock.patch("pcs.lib.tools.get_tmp_file")
        self.addCleanup(tmp_file_patcher.stop)
        tmp_file_mock = tmp_file_patcher.start()
        tmp_file_mock.side_effect = (
            self.tmp_file_mock_obj.get_mock_side_effect()
        )
        self.env_assist, self.config = get_env_tools(test_case=self)

    def config_load_cib_files(self):
        self.config.runner.cib.load(name=self.load_cib_name)
        loaded_cib = self.config.calls.get(self.load_cib_name).stdout
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(self.tmpfile_old, orig_content=loaded_cib),
                TmpFileCall(self.tmpfile_new, orig_content=loaded_cib),
            ]
        )

    def config_load_and_push_diff(self):
        self.config_load_cib_files()
        self.config.runner.cib.diff(self.tmpfile_old, self.tmpfile_new)
        self.config.runner.cib.push_diff()

    def push_reports(self, cib_old=None, cib_new=None):
        # No test changes the CIB between load and push. The point is to test
        # loading and pushing, not editing the CIB.
        loaded_cib = self.config.calls.get(self.load_cib_name).stdout
        return [
            fixture.debug(
                report_codes.TMP_FILE_WRITE,
                file_path=self.tmpfile_old,
                content=(cib_old if cib_old is not None else loaded_cib),
            ),
            fixture.debug(
                report_codes.TMP_FILE_WRITE,
                file_path=self.tmpfile_new,
                content=(
                    cib_new if cib_new is not None else loaded_cib
                ).strip(),
            ),
        ]

    def test_get_and_push(self):
        self.config_load_and_push_diff()
        env = self.env_assist.get_env()

        env.get_cib()
        env.push_cib()
        self.env_assist.assert_reports(self.push_reports())

    def test_can_get_after_push(self):
        self.config_load_and_push_diff()
        self.config.runner.cib.load(name="load_cib_2")
        env = self.env_assist.get_env()

        env.get_cib()
        env.push_cib()
        # need to use lambda because env.cib is a property
        self.assert_raises_cib_not_loaded(lambda: env.cib)
        env.get_cib()
        self.env_assist.assert_reports(self.push_reports())

    def test_not_loaded(self):
        env = self.env_assist.get_env()
        self.assert_raises_cib_not_loaded(env.push_cib)

    def test_tmpfile_fails(self):
        self.config.runner.cib.load()
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(
                    self.tmpfile_old,
                    orig_content=EnvironmentError("test error"),
                ),
            ]
        )
        env = self.env_assist.get_env()

        env.get_cib()
        self.env_assist.assert_raise_library_error(
            env.push_cib,
            [
                fixture.error(
                    report_codes.CIB_SAVE_TMP_ERROR,
                    reason="test error",
                )
            ],
            expected_in_processor=False,
        )

    def test_diff_is_empty(self):
        self.config_load_cib_files()
        self.config.runner.cib.diff(
            self.tmpfile_old,
            self.tmpfile_new,
            stdout="",
            stderr="",
            returncode=1,
        )
        env = self.env_assist.get_env()
        env.get_cib()
        env.push_cib()
        self.env_assist.assert_reports(self.push_reports())

    def test_diff_fails(self):
        self.config_load_cib_files()
        self.config.runner.cib.diff(
            self.tmpfile_old,
            self.tmpfile_new,
            stderr="invalid cib",
            returncode=65,
        )
        loaded_cib = self.config.calls.get(self.load_cib_name).stdout
        env = self.env_assist.get_env()
        env.get_cib()
        self.env_assist.assert_raise_library_error(
            env.push_cib,
            [
                fixture.error(
                    report_codes.CIB_DIFF_ERROR,
                    reason="invalid cib",
                    cib_old=loaded_cib,
                    cib_new=loaded_cib.strip(),
                )
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(self.push_reports())

    def test_push_diff_fails(self):
        self.config_load_cib_files()
        self.config.runner.cib.diff(self.tmpfile_old, self.tmpfile_new)
        self.config.runner.cib.push_diff(stderr="invalid cib", returncode=1)
        env = self.env_assist.get_env()
        env.get_cib()
        self.env_assist.assert_raise_library_error(
            env.push_cib,
            [
                fixture.error(
                    report_codes.CIB_PUSH_ERROR,
                    reason="invalid cib",
                    pushed_cib="",
                )
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(self.push_reports())

    def test_wait(self):
        self.config_load_and_push_diff()
        self.config.runner.pcmk.wait(timeout=self.wait_timeout)
        env = self.env_assist.get_env()

        env.get_cib()
        env.push_cib(wait_timeout=self.wait_timeout)
        self.env_assist.assert_reports(
            self.push_reports()
            + [
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.wait_timeout,
                )
            ]
        )


class PushCustomCib(TestCase, ManageCibAssertionMixin):
    custom_cib = "<custom_cib />"
    wait_timeout = 10

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_push_without_get(self):
        self.config.runner.cib.push_independent(self.custom_cib)
        self.env_assist.get_env().push_cib(etree.XML(self.custom_cib))

    def test_push_after_get(self):
        self.config.runner.cib.load()
        env = self.env_assist.get_env()

        env.get_cib()
        self.assert_raises_cib_loaded_cannot_custom(
            partial(env.push_cib, etree.XML(self.custom_cib))
        )

    def test_wait(self):
        (
            self.config.runner.cib.push_independent(
                self.custom_cib
            ).runner.pcmk.wait(timeout=self.wait_timeout)
        )
        env = self.env_assist.get_env()

        env.push_cib(etree.XML(self.custom_cib), wait_timeout=self.wait_timeout)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.wait_timeout,
                )
            ]
        )


class PushCibMockedWithWait(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_wait_not_supported_for_mocked_cib(self):
        tmp_file = "/fake/tmp/file"
        (
            self.config.env.set_cib_data(
                "<cib/>", cib_tempfile=tmp_file
            ).runner.cib.load(env=dict(CIB_file=tmp_file))
        )

        env = self.env_assist.get_env()
        env.get_cib()
        self.env_assist.assert_raise_library_error(
            lambda: env.push_cib(wait_timeout=10),
            [
                fixture.error(report_codes.WAIT_FOR_IDLE_NOT_LIVE_CLUSTER),
            ],
            expected_in_processor=False,
        )



from pcs import (
    resource,
    stonith,
    usage,
)
from pcs.cli.common.routing import create_router

stonith_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: usage.stonith(argv),
        "list": stonith.stonith_list_available,
        "describe": stonith.stonith_list_options,
        "config": stonith.config_cmd,
        "create": stonith.stonith_create,
        "update": resource.resource_update,
        "update-scsi-devices": stonith.stonith_update_scsi_devices,
        "delete": resource.resource_remove_cmd,
        "remove": resource.resource_remove_cmd,
        # TODO remove, deprecated command
        # replaced with 'stonith status' and 'stonith config'
        "show": stonith.stonith_show_cmd,
        "status": stonith.stonith_status_cmd,
        "level": create_router(
            {
                "add": stonith.stonith_level_add_cmd,
                "clear": stonith.stonith_level_clear_cmd,
                "config": stonith.stonith_level_config_cmd,
                "remove": stonith.stonith_level_remove_cmd,
                "delete": stonith.stonith_level_remove_cmd,
                "verify": stonith.stonith_level_verify_cmd,
            },
            ["stonith", "level"],
            default_cmd="config",
        ),
        "fence": stonith.stonith_fence,
        "cleanup": resource.resource_cleanup,
        "refresh": resource.resource_refresh,
        "confirm": stonith.stonith_confirm,
        "sbd": create_router(
            {
                "enable": stonith.sbd_enable,
                "disable": stonith.sbd_disable,
                "status": stonith.sbd_status,
                "config": stonith.sbd_config,
                "device": create_router(
                    {
                        "setup": stonith.sbd_setup_block_device,
                        "message": stonith.sbd_message,
                    },
                    ["stonith", "sbd", "device"],
                ),
                "watchdog": create_router(
                    {
                        "list": stonith.sbd_watchdog_list,
                        "test": stonith.sbd_watchdog_test,
                        # internal use only
                        "list_json": stonith.sbd_watchdog_list_json,
                    },
                    ["stonith", "sbd", "watchdog"],
                ),
                # internal use only
                "local_config_in_json": stonith.local_sbd_config,
            },
            ["stonith", "sbd"],
        ),
        "enable": resource.resource_enable_cmd,
        "disable": resource.resource_disable_cmd,
        "history": create_router(
            {
                "show": stonith.stonith_history_show_cmd,
                "cleanup": stonith.stonith_history_cleanup_cmd,
                "update": stonith.stonith_history_update_cmd,
            },
            ["stonith", "history"],
            default_cmd="show",
        ),
        # internal use only
        "get_fence_agent_info": stonith.get_fence_agent_info,
    },
    ["stonith"],
    default_cmd="status",
)



from pcs.common import reports
from pcs.common.node_communicator import RequestData
from pcs.common.reports.item import ReportItem
from pcs.lib.communication.tools import (
    AllAtOnceStrategyMixin,
    AllSameDataMixin,
    OneByOneStrategyMixin,
    RunRemotelyBase,
    SimpleResponseProcessingMixin,
    SkipOfflineMixin,
)
from pcs.lib.corosync import live as corosync_live
from pcs.lib.node_communication import response_to_report_item


class Destroy(
    AllSameDataMixin,
    AllAtOnceStrategyMixin,
    SkipOfflineMixin,
    SimpleResponseProcessingMixin,
    RunRemotelyBase,
):
    def _get_request_data(self):
        return RequestData("remote/cluster_destroy")

    def _get_success_report(self, node_label):
        return ReportItem.info(
            reports.messages.ClusterDestroySuccess(node_label)
        )

    def before(self):
        self._set_skip_offline(False, force_code=None)
        self._report(
            ReportItem.info(
                reports.messages.ClusterDestroyStarted(
                    sorted(self._target_label_list)
                )
            )
        )


class DestroyWarnOnFailure(
    AllSameDataMixin, AllAtOnceStrategyMixin, RunRemotelyBase
):
    _unreachable_nodes = None

    def _get_request_data(self):
        return RequestData("remote/cluster_destroy")

    def _process_response(self, response):
        report_item = response_to_report_item(
            response, severity=reports.ReportItemSeverity.WARNING
        )
        node_label = response.request.target.label
        if report_item is None:
            self._report(
                ReportItem.info(
                    reports.messages.ClusterDestroySuccess(node_label)
                )
            )
        else:
            self._report(report_item)
            self._unreachable_nodes.append(node_label)

    def before(self):
        self._report(
            ReportItem.info(
                reports.messages.ClusterDestroyStarted(
                    sorted(self._target_label_list)
                )
            )
        )
        self._unreachable_nodes = []

    def on_complete(self):
        if self._unreachable_nodes:
            self._report(
                ReportItem.warning(
                    reports.messages.NodesToRemoveUnreachable(
                        sorted(self._unreachable_nodes)
                    )
                )
            )


class GetQuorumStatus(AllSameDataMixin, OneByOneStrategyMixin, RunRemotelyBase):
    _quorum_status = None
    _has_failure = False

    def _get_request_data(self):
        return RequestData("remote/get_quorum_info")

    def _process_response(self, response):
        report_item = response_to_report_item(
            response, severity=reports.ReportItemSeverity.WARNING
        )
        node = response.request.target.label
        if report_item is not None:
            self._has_failure = True
            self._report(report_item)
            return self._get_next_list()
        if response.data.strip() == "Cannot initialize CMAP service":
            # corosync is not running on the node, this is OK
            return self._get_next_list()
        try:
            quorum_status = corosync_live.QuorumStatus.from_string(
                response.data
            )
            if not quorum_status.is_quorate:
                return self._get_next_list()
            self._quorum_status = quorum_status
        except corosync_live.QuorumStatusParsingException as e:
            self._has_failure = True
            self._report(
                ReportItem.warning(
                    reports.messages.CorosyncQuorumGetStatusError(
                        e.reason,
                        node=node,
                    )
                )
            )
            return self._get_next_list()
        return []

    def on_complete(self):
        return self._has_failure, self._quorum_status



import json
from unittest import TestCase

from pcs.common.interface.dto import to_dict
from pcs.common.pacemaker.resource.list import ListCibResourcesDto

from pcs_test.tier1.resource.test_config import ResourceConfigCmdMixin
from pcs_test.tools import resources_dto
from pcs_test.tools.misc import get_test_resource
from pcs_test.tools.pcs_runner import PcsRunner


class StonithConfigJson(TestCase):
    def setUp(self):
        self.pcs_runner = PcsRunner(
            cib_file=get_test_resource("cib-resources.xml"),
        )
        self.maxDiff = None

    def test_all(self):
        stdout, retval = self.pcs_runner.run(
            ["stonith", "config", "--output-format=json"],
            ignore_stderr=True,
        )
        self.assertEqual(retval, 0)
        expected = ListCibResourcesDto(
            primitives=[
                resources_dto.STONITH_S2,
                resources_dto.STONITH_S1,
            ],
            clones=[],
            groups=[],
            bundles=[],
        )
        self.assertEqual(json.loads(stdout), to_dict(expected))

    def test_get_specified(self):
        stdout, retval = self.pcs_runner.run(
            ["stonith", "config", "--output-format=json", "S1"],
            ignore_stderr=True,
        )
        self.assertEqual(retval, 0)
        expected = ListCibResourcesDto(
            primitives=[
                resources_dto.STONITH_S1,
            ],
            clones=[],
            groups=[],
            bundles=[],
        )
        self.assertEqual(json.loads(stdout), to_dict(expected))


class StonithConfigCmd(ResourceConfigCmdMixin, TestCase):
    @staticmethod
    def _get_tmp_file_name():
        return "tier1_stonith_test_config_cib.xml"



from typing import (
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    cast,
)

from lxml import etree
from lxml.etree import _Element

from pcs.common import reports
from pcs.common.pacemaker.resource.primitive import CibResourcePrimitiveDto
from pcs.common.resource_agent.dto import ResourceAgentNameDto
from pcs.lib import validate
from pcs.lib.cib import (
    nvpair_multi,
    rule,
)
from pcs.lib.cib.const import TAG_RESOURCE_PRIMITIVE as TAG
from pcs.lib.cib.nvpair import (
    INSTANCE_ATTRIBUTES_TAG,
    append_new_instance_attributes,
    append_new_meta_attributes,
    get_nvset_as_dict,
    get_value,
)
from pcs.lib.cib.resource.agent import get_default_operations
from pcs.lib.cib.resource.operations import (
    create_operations,
    op_element_to_dto,
)
from pcs.lib.cib.resource.operations import prepare as prepare_operations
from pcs.lib.cib.resource.types import ResourceOperationIn
from pcs.lib.cib.tools import (
    IdProvider,
    are_new_role_names_supported,
    does_id_exist,
    find_element_by_tag_and_id,
)
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.pacemaker.live import (
    validate_resource_instance_attributes_via_pcmk,
)
from pcs.lib.pacemaker.values import validate_id
from pcs.lib.resource_agent import (
    ResourceAgentFacade,
    ResourceAgentMetadata,
    ResourceAgentName,
)
from pcs.lib.resource_agent.const import STONITH_ACTION_REPLACED_BY


def is_primitive(resource_el: _Element) -> bool:
    return resource_el.tag == TAG


def primitive_element_to_dto(
    primitive_element: _Element,
    rule_eval: Optional[rule.RuleInEffectEval] = None,
) -> CibResourcePrimitiveDto:
    if rule_eval is None:
        rule_eval = rule.RuleInEffectEvalDummy()
    return CibResourcePrimitiveDto(
        id=str(primitive_element.attrib["id"]),
        agent_name=ResourceAgentNameDto(
            standard=str(primitive_element.attrib["class"]),
            provider=primitive_element.get("provider"),
            type=str(primitive_element.attrib["type"]),
        ),
        description=primitive_element.get("description"),
        operations=[
            op_element_to_dto(op_element, rule_eval)
            for op_element in primitive_element.findall("operations/op")
        ],
        meta_attributes=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                primitive_element, nvpair_multi.NVSET_META
            )
        ],
        instance_attributes=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                primitive_element, nvpair_multi.NVSET_INSTANCE
            )
        ],
        utilization=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                primitive_element, nvpair_multi.NVSET_UTILIZATION
            )
        ],
    )


def _find_primitives_by_agent(
    resources_section: _Element, agent_name: ResourceAgentName
) -> List[_Element]:
    """
    Returns list of primitive resource elements which are using same resource
    agent as specified by resource_agent_obj.

    resources_section -- element <resources/> from CIB
    agent_name -- name of an agent resources of which should be returned
    """
    return cast(
        List[_Element],
        resources_section.xpath(
            ".//primitive[@class=$class_ and @type=$type_ {provider_part}]".format(
                provider_part=" and @provider=$provider_"
                if agent_name.provider
                else "",
            ),
            class_=agent_name.standard,
            provider_=agent_name.provider or "",
            type_=agent_name.type,
        ),
    )


def create(
    report_processor: reports.ReportProcessor,
    cmd_runner: CommandRunner,
    resources_section: _Element,
    id_provider: IdProvider,
    resource_id: str,
    resource_agent_facade: ResourceAgentFacade,
    raw_operation_list: Optional[Iterable[ResourceOperationIn]] = None,
    meta_attributes: Optional[Mapping[str, str]] = None,
    instance_attributes: Optional[Mapping[str, str]] = None,
    allow_invalid_operation: bool = False,
    allow_invalid_instance_attributes: bool = False,
    use_default_operations: bool = True,
    resource_type: str = "resource",
    # TODO remove this arg
    do_not_report_instance_attribute_server_exists: bool = False,
):
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    """
    Prepare all parts of primitive resource and append it into cib.

    report_processor -- a tool for warning/info/error reporting
    resources_section -- a place where the new resource will be appended
    id_provider -- elements' ids generator
    resource_id -- id of the new resource
    resource_agent_facade -- resource agent
    raw_operation_list -- specifies operations of the resource
    meta_attributes -- specifies meta attributes of the resource
    instance_attributes -- specifies instance attributes of the resource
    allow_invalid_operation -- flag for skipping validation of operations
    allow_invalid_instance_attributes -- flag for skipping validation of
        instance_attributes
    use_default_operations -- flag for completion operations with default
        actions specified in resource agent
    resource_type -- describes the resource for reports
    do_not_report_instance_attribute_server_exists -- dirty fix due to
        suboptimal architecture, TODO: fix the architecture and remove the param
    """
    if raw_operation_list is None:
        raw_operation_list = []
    if meta_attributes is None:
        meta_attributes = {}
    if instance_attributes is None:
        instance_attributes = {}

    filtered_raw_operation_list = []
    for op in raw_operation_list:
        filtered_raw_operation_list.append(
            {name: "" if value is None else value for name, value in op.items()}
        )

    if does_id_exist(resources_section, resource_id):
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.IdAlreadyExists(resource_id)
            )
        )
    validate_id(resource_id, "{0} name".format(resource_type))

    agent_metadata = resource_agent_facade.metadata

    operation_list = prepare_operations(
        report_processor,
        filtered_raw_operation_list,
        get_default_operations(
            agent_metadata, necessary_only=not use_default_operations
        ),
        [operation.name for operation in agent_metadata.actions],
        are_new_role_names_supported(resources_section),
        allow_invalid=allow_invalid_operation,
    )

    report_items = validate_resource_instance_attributes_create(
        cmd_runner,
        resource_agent_facade,
        instance_attributes,
        resources_section,
        force=allow_invalid_instance_attributes,
    )
    # TODO remove this "if", see pcs.lib.cib.remote_node.create for details
    if do_not_report_instance_attribute_server_exists:
        for report_item in report_items:
            if isinstance(report_item.message, reports.messages.InvalidOptions):
                report_msg = cast(
                    reports.messages.InvalidOptions, report_item.message
                )
                report_item.message = reports.messages.InvalidOptions(
                    report_msg.option_names,
                    sorted(
                        [
                            value
                            for value in report_msg.allowed
                            if value != "server"
                        ]
                    ),
                    report_msg.option_type,
                    report_msg.allowed_patterns,
                )
    report_processor.report_list(report_items)

    if report_processor.has_errors:
        raise LibraryError()

    return append_new(
        resources_section,
        id_provider,
        resource_id,
        agent_metadata.name.standard,
        agent_metadata.name.provider,
        agent_metadata.name.type,
        instance_attributes=instance_attributes,
        meta_attributes=meta_attributes,
        operation_list=operation_list,
    )


def append_new(
    resources_section,
    id_provider,
    resource_id,
    standard,
    provider,
    agent_type,
    instance_attributes=None,
    meta_attributes=None,
    operation_list=None,
):
    # pylint:disable=too-many-arguments
    """
    Append a new primitive element to the resources_section.

    etree.Element resources_section is place where new element will be appended
    IdProvider id_provider -- elements' ids generator
    string resource_id is id of new resource
    string standard is a standard of resource agent (e.g. ocf)
    string agent_type is a type of resource agent (e.g. IPaddr2)
    string provider is a provider of resource agent (e.g. heartbeat)
    dict instance_attributes will be nvpairs inside instance_attributes element
    dict meta_attributes will be nvpairs inside meta_attributes element
    list operation_list contains dicts representing operations
        (e.g. [{"name": "monitor"}, {"name": "start"}])
    """
    attributes = {
        "id": resource_id,
        "class": standard,
        "type": agent_type,
    }
    if provider:
        attributes["provider"] = provider
    primitive_element = etree.SubElement(resources_section, TAG, attributes)

    if instance_attributes:
        append_new_instance_attributes(
            primitive_element, instance_attributes, id_provider
        )

    if meta_attributes:
        append_new_meta_attributes(
            primitive_element, meta_attributes, id_provider
        )

    create_operations(
        primitive_element, id_provider, operation_list if operation_list else []
    )

    return primitive_element


def _validate_unique_instance_attributes(
    resource_agent: ResourceAgentMetadata,
    instance_attributes: Mapping[str, str],
    resources_section: _Element,
    resource_id: Optional[str] = None,
    force: bool = False,
) -> reports.ReportItemList:
    if not resource_agent.unique_parameter_groups:
        return []

    report_list = []
    same_agent_resources = _find_primitives_by_agent(
        resources_section, resource_agent.name
    )

    for (
        group_name,
        group_attrs,
    ) in resource_agent.unique_parameter_groups.items():
        new_group_values_map = {
            name: instance_attributes.get(name, "") for name in group_attrs
        }
        if not any(new_group_values_map.values()):
            continue

        conflicting_resources: Set[str] = set()
        for primitive in same_agent_resources:
            if primitive.attrib["id"] == resource_id:
                continue
            existing_group_values_map = {
                name: get_value(INSTANCE_ATTRIBUTES_TAG, primitive, name, "")
                for name in group_attrs
            }
            if new_group_values_map == existing_group_values_map:
                conflicting_resources.add(str(primitive.attrib["id"]))

        if conflicting_resources:
            if len(new_group_values_map) == 1:
                message: reports.item.ReportItemMessage = (
                    reports.messages.ResourceInstanceAttrValueNotUnique(
                        *new_group_values_map.popitem(),
                        resource_agent.name.full_name,
                        sorted(conflicting_resources),
                    )
                )
            else:
                message = (
                    reports.messages.ResourceInstanceAttrGroupValueNotUnique(
                        group_name,
                        new_group_values_map,
                        resource_agent.name.full_name,
                        sorted(conflicting_resources),
                    )
                )
            report_list.append(
                reports.ReportItem(
                    reports.item.get_severity(reports.codes.FORCE, force),
                    message,
                )
            )
    return report_list


def _is_ocf_or_stonith_agent(resource_agent_name: ResourceAgentName) -> bool:
    return resource_agent_name.standard in ("stonith", "ocf")


def validate_resource_instance_attributes_create(
    cmd_runner: CommandRunner,
    resource_agent: ResourceAgentFacade,
    instance_attributes: Mapping[str, str],
    resources_section: _Element,
    force: bool = False,
) -> reports.ReportItemList:
    report_items: reports.ReportItemList = []
    agent_name = resource_agent.metadata.name
    if resource_agent.metadata.agent_exists:
        report_items += validate.ValidatorAll(
            resource_agent.get_validators_allowed_parameters(force)
            + resource_agent.get_validators_required_parameters(force)
        ).validate(instance_attributes)
        report_items += validate.ValidatorAll(
            resource_agent.get_validators_deprecated_parameters()
        ).validate(
            {
                name: value
                for name, value in instance_attributes.items()
                # we create a custom report for stonith parameter "action"
                if not (agent_name.is_stonith and name == "action")
            }
        )

    if agent_name.is_stonith:
        report_items += _validate_stonith_action(instance_attributes, force)

    if resource_agent.metadata.agent_exists:
        report_items += _validate_unique_instance_attributes(
            resource_agent.metadata,
            instance_attributes,
            resources_section,
            force=force,
        )

    if (
        _is_ocf_or_stonith_agent(agent_name)
        and resource_agent.metadata.agent_exists
        and resource_agent.metadata.provides_self_validation
        and not any(
            report_item.severity.level == reports.ReportItemSeverity.ERROR
            for report_item in report_items
        )
    ):
        (
            dummy_is_valid,
            agent_validation_reports,
        ) = validate_resource_instance_attributes_via_pcmk(
            cmd_runner,
            agent_name,
            instance_attributes,
            reports.get_severity(reports.codes.FORCE, force),
        )
        report_items.extend(agent_validation_reports)
    return report_items


def validate_resource_instance_attributes_update(
    cmd_runner: CommandRunner,
    resource_agent: ResourceAgentFacade,
    instance_attributes: Mapping[str, str],
    resource_id: str,
    resources_section: _Element,
    force: bool = False,
) -> reports.ReportItemList:
    # pylint: disable=too-many-locals
    # TODO This function currently accepts the updated resource as a string and
    # finds the corresponding xml element by itself. This is needed as the
    # function is called from old pcs code which uses dom while pcs.lib uses
    # lxml. Once resource update command is moved to pcs.lib, this function
    # will be fixed to accept the updated resource as an element instead of a
    # string.
    report_items: reports.ReportItemList = []
    current_instance_attrs = get_nvset_as_dict(
        INSTANCE_ATTRIBUTES_TAG,
        find_element_by_tag_and_id(TAG, resources_section, resource_id),
    )

    agent_name = resource_agent.metadata.name
    if resource_agent.metadata.agent_exists:
        report_items += validate.ValidatorAll(
            resource_agent.get_validators_allowed_parameters(force)
        ).validate(
            # Do not report unknown parameters already set in the CIB. It would
            # be confusing to report an error in an option not actually created
            # now.
            {
                name: value
                for name, value in instance_attributes.items()
                if name not in current_instance_attrs
            }
        )
        report_items += validate.ValidatorAll(
            resource_agent.get_validators_deprecated_parameters()
        ).validate(
            {
                name: value
                for name, value in instance_attributes.items()
                # Allow removing deprecated parameters
                if value != ""
                # we create a custom report for stonith parameter "action"
                and not (agent_name.is_stonith and name == "action")
            }
        )

        # Check that required parameters have not been removed. This is
        # complicated by two facts:
        # * parameters may by deprecated by other parameters, setting one
        #   required parameter from such group is enough
        # * we only want to report errors related to attributes to be updated
        final_attrs = dict(current_instance_attrs)
        for name, value in instance_attributes.items():
            if value == "":
                final_attrs.pop(name, None)
            else:
                final_attrs[name] = value
        report_items += validate.ValidatorAll(
            # Limit validation only to parameters entered now in an update
            # command. We don't want to report missing parameters not mentioned
            # in a command now, that would be confusing to users.
            resource_agent.get_validators_required_parameters(
                force, only_parameters=instance_attributes.keys()
            )
        ).validate(final_attrs)

    if resource_agent.metadata.name.is_stonith:
        report_items += _validate_stonith_action(instance_attributes, force)

    if resource_agent.metadata.agent_exists:
        report_items += _validate_unique_instance_attributes(
            resource_agent.metadata,
            instance_attributes,
            resources_section,
            resource_id=resource_id,
            force=force,
        )

    if (
        _is_ocf_or_stonith_agent(agent_name)
        and resource_agent.metadata.agent_exists
        and resource_agent.metadata.provides_self_validation
        and not any(
            report_item.severity.level == reports.ReportItemSeverity.ERROR
            for report_item in report_items
        )
    ):
        (
            is_valid,
            dummy_reports,
        ) = validate_resource_instance_attributes_via_pcmk(
            cmd_runner,
            agent_name,
            current_instance_attrs,
            reports.ReportItemSeverity.error(),
        )
        if is_valid:
            (
                dummy_is_valid,
                agent_validation_reports,
            ) = validate_resource_instance_attributes_via_pcmk(
                cmd_runner,
                resource_agent.metadata.name,
                final_attrs,
                reports.get_severity(reports.codes.FORCE, force),
            )
            report_items.extend(agent_validation_reports)
    return report_items


def _validate_stonith_action(
    instance_attributes: Mapping[str, str], force: bool = False
) -> reports.ReportItemList:
    if instance_attributes.get("action"):
        return [
            reports.ReportItem(
                reports.item.get_severity(reports.codes.FORCE, force),
                reports.messages.DeprecatedOption(
                    "action", sorted(STONITH_ACTION_REPLACED_BY), "stonith"
                ),
            )
        ]
    return []



from unittest import TestCase

from lxml import etree

from pcs.common.reports import ReportItemSeverity as severities
from pcs.common.reports import codes as report_codes
from pcs.lib.cib import sections

from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_xml_equal,
)
from pcs_test.tools.xml import etree_to_str


class Get(TestCase):
    def setUp(self):
        self.tree = etree.fromstring(
            """
            <cib>
                <configuration>
                    <acls/>
                </configuration>
            </cib>
            """
        )

    @staticmethod
    def assert_element_content(section_element, expected_xml):
        assert_xml_equal(etree_to_str(section_element), expected_xml)

    def test_get_existing_mandatory(self):
        self.assert_element_content(
            sections.get(self.tree, sections.CONFIGURATION),
            """
            <configuration>
                <acls/>
            </configuration>
            """,
        )

    def test_get_existing_optional(self):
        self.assert_element_content(
            sections.get(self.tree, sections.ACLS), "<acls/>"
        )

    def test_get_no_existing_optional(self):
        self.assert_element_content(
            sections.get(self.tree, sections.ALERTS), "<alerts/>"
        )
        self.assert_element_content(
            self.tree,
            """
            <cib>
                <configuration>
                    <acls/>
                    <alerts/>
                </configuration>
            </cib>
            """,
        )

    def test_raises_on_no_existing_mandatory_section(self):
        assert_raise_library_error(
            lambda: sections.get(self.tree, sections.NODES),
            (
                severities.ERROR,
                report_codes.CIB_CANNOT_FIND_MANDATORY_SECTION,
                {
                    "section": "configuration/nodes",
                },
            ),
        )



from unittest import TestCase


class GetCaCert(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.AddDeviceNetTest
    """


class ClientSetup(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.AddDeviceNetTest
    """


class SignCertificate(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.AddDeviceNetTest
    """


class ClientImportCertificateAndKey(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.AddDeviceNetTest
    """


class ClientDestroy(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.test_quorum.RemoveDeviceNetTest
    """



from typing import Sequence

from ..types import ExecutorResult


class ExecutorInterface:
    """
    Simple interface for executing external programs.
    """

    def run(self, args: Sequence[str]) -> ExecutorResult:
        """
        args -- Program and its arguments to execute. First item is path to a
            executable and rest of the items are arguments which will be provided
            to the executable.

        Execute a specified program synchronously and return its result after
        it's finished.
        """
        raise NotImplementedError()



from collections import namedtuple
from unittest import TestCase

from lxml import etree

from pcs.common import reports
from pcs.lib.cib import tag as lib
from pcs.lib.cib.tools import (
    IdProvider,
    get_constraints,
    get_resources,
    get_tags,
)
from pcs.lib.xml_tools import etree_to_str

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
    assert_xml_equal,
)

# pylint: disable=protected-access
# pylint: disable=too-many-lines

TagWithRefs = namedtuple("TagWithRefs", ["tag_id", "idref_list"])

FIXTURE_ONE_TAG = """
    <cib>
        <configuration>
            <tags>
                <tag id="first_tag">
                    <obj_ref id="idref1"/>
                    <obj_ref id="idref2"/>
                </tag>
            </tags>
        </configuration>
    </cib>
"""

FIXTURE_TWO_TAGS = """
    <cib>
        <configuration>
            <tags>
                <tag id="first_tag">
                    <obj_ref id="idref1"/>
                    <obj_ref id="idref2"/>
                </tag>
                <tag id="second_tag">
                    <obj_ref id="ref1"/>
                    <obj_ref id="ref2"/>
                    <obj_ref id="ref3"/>
                </tag>
            </tags>
        </configuration>
    </cib>
"""


FIXTURE_OBJ_REFS = """
    <tags>
        <tag id="tag-A">
            <obj_ref id="ref1"/>
            <obj_ref id="ref2"/>
            <obj_ref id="ref3"/>
        </tag>
        <tag id="tag-B">
            <obj_ref id="common"/>
            <obj_ref id="ref4"/>
            <obj_ref id="ref5"/>
        </tag>
        <tag id="tag-C">
            <obj_ref id="common"/>
            <obj_ref id="ref4"/>
        </tag>
    </tags>
    """


def get_elements_by_ids(parent, id_list):
    element_list = []
    for _id in id_list:
        found_elements_list = parent.findall(
            f'.//*[@id="{_id}"]',
        )
        if found_elements_list:
            element_list.extend(found_elements_list)
    return element_list


def fixture_unexpected_element_reports(id_list, expected_types=None):
    return [
        fixture.report_unexpected_element(
            _id,
            ValidateCommonTestData.id_to_context_type_map[_id],
            expected_types=[
                "bundle",
                "clone",
                "group",
                "master",
                "primitive",
            ]
            if expected_types is None
            else expected_types,
        )
        for _id in id_list
    ]


def fixture_id_not_found_reports(id_list, expected_types=None):
    return [
        fixture.report_not_found(
            _id,
            context_type=ValidateCommonTestData.id_to_context_type_map[_id],
            expected_types=expected_types,
        )
        for _id in id_list
    ]


class LibraryTagTest(TestCase):
    tag_sets = [
        TagWithRefs("first_tag", ["idref1", "idref2"]),
        TagWithRefs("second_tag", ["ref1", "ref2", "ref3"]),
    ]

    def setUp(self):
        self.cib = etree.fromstring("<cib><configuration/></cib>")

    def call_create_tag(self, tag_with_refs):
        lib.create_tag(
            get_tags(self.cib),
            tag_with_refs.tag_id,
            tag_with_refs.idref_list,
        )

    def assert_cib_equal(self, expected_cib):
        assert_xml_equal(expected_cib, etree_to_str(self.cib))

    def test_create_tag_if_no_tags(self):
        self.call_create_tag(self.tag_sets[0])
        self.assert_cib_equal(FIXTURE_ONE_TAG)

    def test_create_tag_if_tags_exists(self):
        self.cib = etree.fromstring(FIXTURE_ONE_TAG)
        self.call_create_tag(self.tag_sets[1])
        self.assert_cib_equal(FIXTURE_TWO_TAGS)


class ValidateCommonTestData(TestCase):
    tag_id = "tag_id"
    resource_ids = ["id-01", "id-02", "id-03"]
    nonresource_ids = ["no1", "no2"]
    nonexistent_ids = ["nonexistent1", "nonexistent2"]
    nonexistent_tags = ["nonexistent_tag1", "nonexistent_tag2"]
    existent_tags = ["first_tag", "second_tag"]
    id_to_context_type_map = {
        "no1": "rsc_location",
        "no2": "resources",
        "nonexistent1": "resources",
        "nonexistent2": "resources",
        "#invalid-tag-id": "resources",
        "tag_id": "resources",
        "nonexistent_tag1": "tags",
        "nonexistent_tag2": "tags",
    }
    test_tree = etree.fromstring(
        """
        <cib>
          <configuration>
            <resources>
              <primitive id="{resource_ids[0]}"/>
              <group id="G">
                <primitive id="{resource_ids[1]}"/>
              </group>
              <clone id="C">
                <primitive id="{resource_ids[2]}"/>
              </clone>
            </resources>
            <constraints>
              <rsc_location id="{nonresource_ids[0]}"/>
            </constraints>
            <tags>
              <tag id="{tags[0]}">
                <obj_ref id="idref1"/>
                <obj_ref id="idref2"/>
              </tag>
              <tag id="{tags[1]}">
                <obj_ref id="ref1"/>
                <obj_ref id="ref2"/>
                <obj_ref id="ref3"/>
              </tag>
            </tags>
          </configuration>
          <status>
            <node_state id="{nonresource_ids[1]}"/>
          </status>
        </cib>
        """.format(
            resource_ids=resource_ids,
            tags=existent_tags,
            nonresource_ids=nonresource_ids,
        )
    )
    tags_section = get_tags(test_tree)
    resources_section = get_resources(test_tree)
    id_provider = IdProvider(tags_section)

    def get_tag_elements(self, id_list):
        return [
            element
            for element in self.tags_section.findall("tag")
            if element.get("id", default="") in id_list
        ]


class AllTagElementsToDict(ValidateCommonTestData):
    def test_success(self):
        element_list = lib.get_list_of_tag_elements(get_tags(self.test_tree))
        self.assertEqual(
            [lib.tag_element_to_dict(element) for element in element_list],
            [
                {
                    "tag_id": "first_tag",
                    "idref_list": ["idref1", "idref2"],
                },
                {
                    "tag_id": "second_tag",
                    "idref_list": ["ref1", "ref2", "ref3"],
                },
            ],
        )


class ValidateCreateTag(ValidateCommonTestData):
    def test_validation_success(self):
        assert_report_item_list_equal(
            lib.validate_create_tag(
                get_resources(self.test_tree),
                "new-tag",
                self.resource_ids,
                self.id_provider,
            ),
            [],
        )

    def test_validation_failure_all_kinds_reports(self):
        tag_id = "#invalid-tag-id"
        idref_list = 2 * self.nonresource_ids
        idref_list.append(tag_id)
        assert_report_item_list_equal(
            lib.validate_create_tag(
                get_resources(self.test_tree),
                tag_id,
                idref_list,
                self.id_provider,
            ),
            [
                fixture.report_invalid_id(tag_id, "#", id_description="id"),
                fixture.error(reports.codes.TAG_CANNOT_CONTAIN_ITSELF),
                *fixture_unexpected_element_reports(
                    2 * self.nonresource_ids[:1]
                ),
                *fixture_id_not_found_reports(2 * self.nonresource_ids[1:]),
                *fixture_id_not_found_reports([tag_id]),
                fixture.error(
                    reports.codes.TAG_ADD_REMOVE_IDS_DUPLICATION,
                    duplicate_ids_list=self.nonresource_ids,
                    add_or_not_remove=True,
                ),
            ],
        )


class ValidateTagIdTest(ValidateCommonTestData):
    def test_tag_id_is_valid(self):
        assert_report_item_list_equal(
            lib._validate_tag_id("new-tag-id", self.id_provider),
            [],
        )

    def test_tag_id_is_empty(self):
        assert_report_item_list_equal(
            lib._validate_tag_id("", self.id_provider),
            [
                fixture.error(
                    reports.codes.INVALID_ID_IS_EMPTY,
                    id_description="id",
                )
            ],
        )

    def test_tag_id_invalid_first_character(self):
        assert_report_item_list_equal(
            lib._validate_tag_id("0tagid", self.id_provider),
            [fixture.report_invalid_id("0tagid", "0")],
        )
        assert_report_item_list_equal(
            lib._validate_tag_id("-tagid", self.id_provider),
            [fixture.report_invalid_id("-tagid", "-")],
        )
        assert_report_item_list_equal(
            lib._validate_tag_id("átagid", self.id_provider),
            [fixture.report_invalid_id("átagid", "á")],
        )

    def test_tag_id_invalid_character(self):
        assert_report_item_list_equal(
            lib._validate_tag_id("tag,id", self.id_provider),
            [fixture.report_invalid_id("tag,id", ",")],
        )
        assert_report_item_list_equal(
            lib._validate_tag_id("tagid>", self.id_provider),
            [fixture.report_invalid_id("tagid>", ">")],
        )
        assert_report_item_list_equal(
            lib._validate_tag_id("taígid", self.id_provider),
            [fixture.report_invalid_id("taígid", "í")],
        )

    def test_tag_id_already_exists(self):
        for tag_id in ["id-01", "id-02"]:
            assert_report_item_list_equal(
                lib._validate_tag_id(tag_id, self.id_provider),
                [fixture.error(reports.codes.ID_ALREADY_EXISTS, id=tag_id)],
            )


class ValidateTagIdNotInIdrefList(TestCase):
    def setUp(self):
        self.idref_list = ["id1", "id2", "id3"]

    def test_tag_contains_itself(self):
        assert_report_item_list_equal(
            lib._validate_tag_id_not_in_idref_list(
                self.idref_list[0],
                self.idref_list,
            ),
            [fixture.error(reports.codes.TAG_CANNOT_CONTAIN_ITSELF)],
        )

    def test_tag_does_not_contain_itself(self):
        assert_report_item_list_equal(
            lib._validate_tag_id_not_in_idref_list(
                "tag-not-in-list",
                self.idref_list,
            ),
            [],
        )


class ValidateDuplicateReferenceIds(ValidateCommonTestData):
    duplicated_ids_input_output = (
        (
            ["id1", "id1"],
            ["id1"],
        ),
        (
            ["id1", "id2", "id2"],
            ["id2"],
        ),
        (
            ["id1", "id2", "id3", "id1", "id4", "id4"],
            ["id1", "id4"],
        ),
    )

    def test_no_duplicates(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            lib._validate_add_remove_duplicate_reference_ids(["id1", "id2"]),
            [],
        )

    def test_duplicates(self):
        # pylint: disable=no-self-use
        for input_ids, output_ids in self.duplicated_ids_input_output:
            assert_report_item_list_equal(
                lib._validate_add_remove_duplicate_reference_ids(input_ids),
                [
                    fixture.error(
                        reports.codes.TAG_ADD_REMOVE_IDS_DUPLICATION,
                        duplicate_ids_list=output_ids,
                        add_or_not_remove=True,
                    )
                ],
            )


class ValidateTagCreateIdrefListNotEmpty(TestCase):
    def test_empty_list(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            lib._validate_tag_create_idref_list_not_empty([]),
            [
                fixture.error(
                    reports.codes.TAG_CANNOT_CREATE_EMPTY_TAG_NO_IDS_SPECIFIED,
                )
            ],
        )

    def test_not_empty_list(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            lib._validate_tag_create_idref_list_not_empty(["id"]),
            [],
        )


class ValidateReferenceIdsAreResources(ValidateCommonTestData):
    def test_ids_exist(self):
        assert_report_item_list_equal(
            lib._validate_reference_ids_are_resources(
                get_resources(self.test_tree),
                self.resource_ids,
            ),
            [],
        )

    def test_no_ids_empty_list(self):
        assert_report_item_list_equal(
            lib._validate_reference_ids_are_resources(
                get_resources(self.test_tree),
                [],
            ),
            [],
        )

    def test_id_does_not_exist(self):
        assert_report_item_list_equal(
            lib._validate_reference_ids_are_resources(
                get_resources(self.test_tree),
                self.nonexistent_ids[0:1],
            ),
            fixture_id_not_found_reports(self.nonexistent_ids[0:1]),
        )

    def test_multiple_id_does_not_exist(self):
        assert_report_item_list_equal(
            lib._validate_reference_ids_are_resources(
                get_resources(self.test_tree),
                self.nonexistent_ids,
            ),
            fixture_id_not_found_reports(self.nonexistent_ids),
        )

    def test_id_is_not_a_resource(self):
        assert_report_item_list_equal(
            lib._validate_reference_ids_are_resources(
                get_resources(self.test_tree),
                self.nonresource_ids[0:1],
            ),
            fixture_unexpected_element_reports(self.nonresource_ids[0:1]),
        )

    def test_ids_are_not_resources(self):
        assert_report_item_list_equal(
            lib._validate_reference_ids_are_resources(
                get_resources(self.test_tree),
                self.nonresource_ids,
            ),
            fixture_unexpected_element_reports(self.nonresource_ids[:1])
            + fixture_id_not_found_reports(self.nonresource_ids[1:]),
        )

    def test_mixed_ids(self):
        assert_report_item_list_equal(
            lib._validate_reference_ids_are_resources(
                get_resources(self.test_tree),
                self.resource_ids + self.nonresource_ids + self.nonexistent_ids,
            ),
            fixture_unexpected_element_reports(self.nonresource_ids[:1])
            + fixture_id_not_found_reports(self.nonresource_ids[1:])
            + fixture_id_not_found_reports(self.nonexistent_ids),
        )


class FindTagElementsByIdsTest(ValidateCommonTestData):
    def call_find_elements_by_ids(self, id_list):
        return lib.find_tag_elements_by_ids(
            self.tags_section,
            id_list,
        )

    def test_ids_exist(self):
        element_list, report_list = self.call_find_elements_by_ids(
            self.existent_tags,
        )
        assert_report_item_list_equal(
            report_list,
            [],
        )
        self.assertEqual(
            element_list,
            self.get_tag_elements(self.existent_tags),
        )

    def test_ids_does_not_exist(self):
        element_list, report_list = self.call_find_elements_by_ids(
            self.nonexistent_tags,
        )
        assert_report_item_list_equal(
            report_list,
            fixture_id_not_found_reports(
                self.nonexistent_tags,
                expected_types=["tag"],
            ),
        )
        self.assertEqual(element_list, [])

    def test_mixed_ids(self):
        element_list, report_list = self.call_find_elements_by_ids(
            self.existent_tags[0:1] + self.nonexistent_tags,
        )
        assert_report_item_list_equal(
            report_list,
            fixture_id_not_found_reports(
                self.nonexistent_tags,
                expected_types=["tag"],
            ),
        )
        self.assertEqual(
            element_list,
            self.get_tag_elements(self.existent_tags[0:1]),
        )

    def test_not_tag_ids(self):
        element_list, report_list = self.call_find_elements_by_ids(
            self.nonresource_ids,
        )
        assert_report_item_list_equal(
            report_list,
            fixture_unexpected_element_reports(
                self.nonresource_ids[:1], expected_types=["tag"]
            )
            + [
                fixture.report_not_found(
                    self.nonresource_ids[1],
                    context_type="tags",
                    expected_types=["tag"],
                )
            ],
        )
        self.assertEqual(element_list, [])


class LibraryRemoveTag(TestCase):
    def setUp(self):
        self.cib = etree.fromstring(FIXTURE_TWO_TAGS)

    def test_remove_empty_list(self):
        lib.remove_tag([])
        assert_xml_equal(FIXTURE_TWO_TAGS, etree_to_str(self.cib))

    def test_remove_a_tag(self):
        lib.remove_tag(
            elem
            for elem in get_tags(self.cib)
            if elem.get("id") == "second_tag"
        )
        assert_xml_equal(FIXTURE_ONE_TAG, etree_to_str(self.cib))

    def test_remove_more_tags(self):
        lib.remove_tag(elem for elem in get_tags(self.cib))
        assert_xml_equal(
            "<cib><configuration><tags/></configuration></cib>",
            etree_to_str(self.cib),
        )


class ValidateCommonConstraintsTestData(TestCase):
    constraint_tags = [
        "tag-colocation",
        "tag-colocation-with",
        "tag-location",
        "tag-order-first",
        "tag-order-then",
        "tag-ticket",
        "tag-colocation-set",
        "tag-location-set",
        "tag-order-set",
        "tag-ticket-set",
    ]
    tag2constraint_id = {
        "tag-colocation": "colocation",
        "tag-colocation-with": "colocation",
        "tag-location": "location",
        "tag-order-first": "order",
        "tag-order-then": "order",
        "tag-ticket": "ticket",
        "tag-colocation-set": "colocation-set",
        "tag-location-set": "location-set",
        "tag-order-set": "order-set",
        "tag-ticket-set": "ticket-set",
    }
    constraint_template = """
        <cib>
          <configuration>
            <resources/>
            <constraints>

              <rsc_colocation id="colocation"
                rsc="{tags[0]}"
                with-rsc="{tags[1]}"/>
              <rsc_location id="location" rsc="{tags[2]}"/>
              <rsc_order id="order" first="{tags[3]}" then="{tags[4]}"/>
              <rsc_ticket id="ticket" rsc="{tags[5]}"/>

              <rsc_colocation id="colocation-set">
                <resource_set id="in-colocation-set">
                  <resource_ref id="{tags[6]}"/>
                  <resource_ref id="dummy"/>
                </resource_set>
              </rsc_colocation>
              <rsc_location id="location-set">
                <resource_set id="in-location-set">
                  <resource_ref id="{tags[7]}"/>
                  <resource_ref id="dummy"/>
                </resource_set>
              </rsc_location>
              <rsc_order id="order-set">
                <resource_set id="in-order-set">
                  <resource_ref id="{tags[8]}"/>
                  <resource_ref id="dummy"/>
                </resource_set>
              </rsc_order>
              <rsc_ticket id="ticket-set">
                <resource_set id="in-ticket-set">
                  <resource_ref id="{tags[9]}"/>
                  <resource_ref id="dummy"/>
                </resource_set>

              </rsc_ticket>
            </constraints>
          </configuration>
          <status/>
        </cib>
        """
    tree_each_tag_has_one_constraint = etree.fromstring(
        constraint_template.format(tags=constraint_tags)
    )
    tree_tag_has_multiple_constraints = etree.fromstring(
        constraint_template.format(tags=["multitag"] * len(constraint_tags))
    )


class ValidateRemoveTag(ValidateCommonConstraintsTestData):
    def test_success_non_empty_list_for_remove(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            lib.validate_remove_tag(
                get_constraints(self.tree_each_tag_has_one_constraint),
                ["not-referenced-tag"],
            ),
            [],
        )

    def test_fail_empty_list_for_remove(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            lib.validate_remove_tag(
                get_constraints(self.tree_each_tag_has_one_constraint),
                [],
            ),
            [
                fixture.error(
                    reports.codes.TAG_CANNOT_REMOVE_TAGS_NO_TAGS_SPECIFIED
                )
            ],
        )

    def test_fail_tag_referenced_in_constraint(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            lib.validate_remove_tag(
                get_constraints(self.tree_each_tag_has_one_constraint),
                ["tag-location"],
            ),
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_REMOVE_TAG_REFERENCED_IN_CONSTRAINTS,
                    tag_id="tag-location",
                    constraint_id_list=["location"],
                )
            ],
        )

    def test_fail_tag_referenced_in_multiple_constraint(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            lib.validate_remove_tag(
                get_constraints(self.tree_tag_has_multiple_constraints),
                ["multitag"],
            ),
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_REMOVE_TAG_REFERENCED_IN_CONSTRAINTS,
                    tag_id="multitag",
                    constraint_id_list=sorted(
                        set(self.tag2constraint_id.values()),
                    ),
                )
            ],
        )


class FindConstraintsReferencingTag(ValidateCommonConstraintsTestData):
    @staticmethod
    def call_find_constraints_referencing_tag(tree, tag_id):
        return lib.find_constraints_referencing_tag(
            get_constraints(tree),
            tag_id,
        )

    def assert_constraint_id(self, tag_id):
        one_constraint_list = self.call_find_constraints_referencing_tag(
            self.tree_each_tag_has_one_constraint,
            tag_id,
        )
        self.assertEqual(len(one_constraint_list), 1)
        self.assertEqual(
            one_constraint_list[0].get("id"),
            self.tag2constraint_id[tag_id],
        )

    def test_multiple_constraints(self):
        constraint_list = self.call_find_constraints_referencing_tag(
            self.tree_tag_has_multiple_constraints,
            "multitag",
        )
        constraint_id_set = set(self.tag2constraint_id.values())
        self.assertEqual(
            len(constraint_list),
            len(constraint_id_set),
        )
        for constraint in constraint_list:
            self.assertIn(constraint.get("id"), constraint_id_set)

    def test_colocation(self):
        self.assert_constraint_id("tag-colocation")

    def test_colocation_with(self):
        self.assert_constraint_id("tag-colocation-with")

    def test_location(self):
        self.assert_constraint_id("tag-location")

    def test_order_first(self):
        self.assert_constraint_id("tag-order-first")

    def test_order_then(self):
        self.assert_constraint_id("tag-order-then")

    def test_ticket(self):
        self.assert_constraint_id("tag-ticket")

    def test_colocation_set(self):
        self.assert_constraint_id("tag-colocation-set")

    def test_location_set(self):
        self.assert_constraint_id("tag-location-set")

    def test_order_set(self):
        self.assert_constraint_id("tag-order-set")

    def test_ticket_set(self):
        self.assert_constraint_id("tag-ticket-set")


class ValidateAddRemoveDuplicateReferenceIds(ValidateCommonTestData):
    duplicated_ids_input_output = (
        (
            ["id1", "id1"],
            ["id1"],
        ),
        (
            ["id1", "id2", "id2"],
            ["id2"],
        ),
        (
            ["id1", "id2", "id3", "id1", "id4", "id4"],
            ["id1", "id4"],
        ),
    )

    def test_add_no_duplicates(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            lib._validate_add_remove_duplicate_reference_ids(["id1", "id2"]),
            [],
        )

    def test_remove_no_duplicates(self):
        # pylint: disable=no-self-use
        assert_report_item_list_equal(
            lib._validate_add_remove_duplicate_reference_ids(
                ["id1", "id2"],
                False,
            ),
            [],
        )

    def test_add_duplicates(self):
        for input_ids, output_ids in self.duplicated_ids_input_output:
            with self.subTest(input_ids=input_ids, output_ids=output_ids):
                assert_report_item_list_equal(
                    lib._validate_add_remove_duplicate_reference_ids(input_ids),
                    [
                        fixture.error(
                            reports.codes.TAG_ADD_REMOVE_IDS_DUPLICATION,
                            duplicate_ids_list=output_ids,
                            add_or_not_remove=True,
                        )
                    ],
                )

    def test_remove_duplicates(self):
        for input_ids, output_ids in self.duplicated_ids_input_output:
            with self.subTest(input_ids=input_ids, output_ids=output_ids):
                assert_report_item_list_equal(
                    lib._validate_add_remove_duplicate_reference_ids(
                        input_ids,
                        False,
                    ),
                    [
                        fixture.error(
                            reports.codes.TAG_ADD_REMOVE_IDS_DUPLICATION,
                            duplicate_ids_list=output_ids,
                            add_or_not_remove=False,
                        )
                    ],
                )


class LibraryRemoveObjRef(TestCase):
    def setUp(self):
        self.cib = etree.fromstring(FIXTURE_OBJ_REFS)

    def test_remove_empty_list(self):
        lib.remove_obj_ref([])
        assert_xml_equal(FIXTURE_OBJ_REFS, etree_to_str(self.cib))

    def test_remove_refs_from_a_tag_without_removing_tag(self):
        lib.remove_obj_ref(get_elements_by_ids(self.cib, ["ref1", "ref2"]))
        assert_xml_equal(
            """
            <tags>
                <tag id="tag-A">
                    <obj_ref id="ref3"/>
                </tag>
                <tag id="tag-B">
                    <obj_ref id="common"/>
                    <obj_ref id="ref4"/>
                    <obj_ref id="ref5"/>
                </tag>
                <tag id="tag-C">
                    <obj_ref id="common"/>
                    <obj_ref id="ref4"/>
                </tag>
            </tags>
            """,
            etree_to_str(self.cib),
        )

    def test_remove_refs_from_a_tag_with_removing_tag(self):
        lib.remove_obj_ref(
            get_elements_by_ids(self.cib, ["ref1", "ref2", "ref3"])
        )
        assert_xml_equal(
            """
            <tags>
                <tag id="tag-B">
                    <obj_ref id="common"/>
                    <obj_ref id="ref4"/>
                    <obj_ref id="ref5"/>
                </tag>
                <tag id="tag-C">
                    <obj_ref id="common"/>
                    <obj_ref id="ref4"/>
                </tag>
            </tags>
            """,
            etree_to_str(self.cib),
        )

    def test_remove_refs_from_multiple_tags_without_removing_tag(self):
        lib.remove_obj_ref(get_elements_by_ids(self.cib, ["ref4", "ref5"]))
        assert_xml_equal(
            """
            <tags>
                <tag id="tag-A">
                    <obj_ref id="ref1"/>
                    <obj_ref id="ref2"/>
                    <obj_ref id="ref3"/>
                </tag>
                <tag id="tag-B">
                    <obj_ref id="common"/>
                </tag>
                <tag id="tag-C">
                    <obj_ref id="common"/>
                </tag>
            </tags>
            """,
            etree_to_str(self.cib),
        )

    def test_remove_refs_from_multiple_tags_with_removing_tags(self):
        lib.remove_obj_ref(get_elements_by_ids(self.cib, ["common", "ref4"]))
        assert_xml_equal(
            """
            <tags>
                <tag id="tag-A">
                    <obj_ref id="ref1"/>
                    <obj_ref id="ref2"/>
                    <obj_ref id="ref3"/>
                </tag>
                <tag id="tag-B">
                    <obj_ref id="ref5"/>
                </tag>
            </tags>
            """,
            etree_to_str(self.cib),
        )


class LibraryAddObjRef(TestCase):
    def setUp(self):
        self.tag_el = etree.fromstring(
            """
            <tag>
                <obj_ref id="e1"/>
                <obj_ref id="e2"/>
                <obj_ref id="e3"/>
            </tag>
            """
        )
        self.new_list = [etree.Element("obj_ref", id=_id) for _id in "ab"]
        self.existing_list = get_elements_by_ids(
            self.tag_el,
            ["e1", "e2", "e3"],
        )

    def test_add_new_ids(self):
        lib.add_obj_ref(
            self.tag_el,
            self.new_list,
            None,
        )
        assert_xml_equal(
            """
            <tag>
                <obj_ref id="e1"/>
                <obj_ref id="e2"/>
                <obj_ref id="e3"/>
                <obj_ref id="a"/>
                <obj_ref id="b"/>
            </tag>
            """,
            etree_to_str(self.tag_el),
        )

    def test_add_new_ids_before_adjacent(self):
        lib.add_obj_ref(
            self.tag_el,
            self.new_list,
            self.existing_list[0],
        )
        assert_xml_equal(
            """
            <tag>
                <obj_ref id="a"/>
                <obj_ref id="b"/>
                <obj_ref id="e1"/>
                <obj_ref id="e2"/>
                <obj_ref id="e3"/>
            </tag>
            """,
            etree_to_str(self.tag_el),
        )

    def test_add_new_ids_after_adjacent(self):
        lib.add_obj_ref(
            self.tag_el,
            self.new_list,
            self.existing_list[1],
            put_after_adjacent=True,
        )
        assert_xml_equal(
            """
            <tag>
                <obj_ref id="e1"/>
                <obj_ref id="e2"/>
                <obj_ref id="a"/>
                <obj_ref id="b"/>
                <obj_ref id="e3"/>
            </tag>
            """,
            etree_to_str(self.tag_el),
        )

    def test_move_existing_ids_before(self):
        lib.add_obj_ref(
            self.tag_el,
            self.existing_list[2:0:-1],
            self.existing_list[0],
        )
        assert_xml_equal(
            """
            <tag>
                <obj_ref id="e3"/>
                <obj_ref id="e2"/>
                <obj_ref id="e1"/>
            </tag>
            """,
            etree_to_str(self.tag_el),
        )

    def test_move_existing_ids_after(self):
        lib.add_obj_ref(
            self.tag_el,
            self.existing_list[2::-2],
            self.existing_list[1],
            put_after_adjacent=True,
        )
        assert_xml_equal(
            """
            <tag>
                <obj_ref id="e2"/>
                <obj_ref id="e3"/>
                <obj_ref id="e1"/>
            </tag>
            """,
            etree_to_str(self.tag_el),
        )

    def test_move_both_new_and_existing_before(self):
        el_list = [
            self.new_list[1],
            self.existing_list[0],
            self.new_list[0],
            self.existing_list[2],
        ]
        lib.add_obj_ref(self.tag_el, el_list, self.existing_list[1])
        assert_xml_equal(
            """
            <tag>
                <obj_ref id="b"/>
                <obj_ref id="e1"/>
                <obj_ref id="a"/>
                <obj_ref id="e3"/>
                <obj_ref id="e2"/>
            </tag>
            """,
            etree_to_str(self.tag_el),
        )

    def test_move_both_new_and_existing_after(self):
        el_list = [
            self.new_list[1],
            self.new_list[0],
            self.existing_list[0],
            self.existing_list[1],
        ]
        lib.add_obj_ref(
            self.tag_el,
            el_list,
            self.existing_list[2],
            put_after_adjacent=True,
        )
        assert_xml_equal(
            """
            <tag>
                <obj_ref id="e3"/>
                <obj_ref id="b"/>
                <obj_ref id="a"/>
                <obj_ref id="e1"/>
                <obj_ref id="e2"/>
            </tag>
            """,
            etree_to_str(self.tag_el),
        )


class ValidateTagUpdateByIds(TestCase):
    OBJ_REF_EXPECTED_TYPES = ["bundle", "clone", "group", "master", "primitive"]
    test_cib = etree.fromstring(
        """
        <cib>
          <configuration>
            <resources>
              <primitive id="e1"/>
              <primitive id="e2"/>
              <primitive id="e3"/>
              <primitive id="new1"/>
            </resources>
            <tags>
              <tag id="tag">
                <obj_ref id="e1"/>
                <obj_ref id="e2"/>
                <obj_ref id="e3"/>
              </tag>
            </tags>
            <rsc_defaults>
                <meta_attributes id="other_id" />
            </rsc_defaults>
          </configuration>
          <status>
            <node_state id="status_id"/>
          </status>
        </cib>
        """
    )

    def _validate(self, tag, to_add, to_remove, adjacent=None):
        return lib.ValidateTagUpdateByIds(
            tag,
            to_add,
            to_remove,
            adjacent_idref=adjacent,
        ).validate(get_resources(self.test_cib), get_tags(self.test_cib))

    def test_tag_id_not_a_tag(self):
        assert_report_item_list_equal(
            self._validate("new1", ["new1"], ["none"], adjacent="none"),
            [fixture.report_unexpected_element("new1", "primitive", ["tag"])],
        )

    def test_not_existing_tag(self):
        assert_report_item_list_equal(
            self._validate("none_tag", ["e1"], []),
            [
                fixture.report_not_found(
                    "none_tag",
                    context_type="tags",
                    expected_types=["tag"],
                ),
            ],
        )

    def test_not_existing_tag_and_no_ids(self):
        assert_report_item_list_equal(
            self._validate("none_tag", [], []),
            [
                fixture.report_not_found(
                    "none_tag",
                    context_type="tags",
                    expected_types=["tag"],
                ),
                fixture.error(
                    reports.codes.TAG_CANNOT_UPDATE_TAG_NO_IDS_SPECIFIED,
                ),
            ],
        )

    def test_not_existing_tag_id_and_bad_add_ids(self):
        assert_report_item_list_equal(
            self._validate("none_tag", ["none", "status_id", "other_id"], []),
            [
                fixture.report_not_found(
                    "none_tag",
                    context_type="tags",
                    expected_types=["tag"],
                ),
                fixture.report_not_found(
                    "none",
                    context_type="resources",
                    expected_types=self.OBJ_REF_EXPECTED_TYPES,
                ),
                fixture.report_not_found(
                    "status_id",
                    context_type="resources",
                    expected_types=self.OBJ_REF_EXPECTED_TYPES,
                ),
                fixture.report_unexpected_element(
                    "other_id", "meta_attributes", self.OBJ_REF_EXPECTED_TYPES
                ),
            ],
        )

    def test_existing_tag_and_no_ids(self):
        assert_report_item_list_equal(
            self._validate("tag", [], []),
            [
                fixture.error(
                    reports.codes.TAG_CANNOT_UPDATE_TAG_NO_IDS_SPECIFIED,
                ),
            ],
        )

    def test_existing_tag_and_no_ids_with_not_existing_adjacent(self):
        assert_report_item_list_equal(
            self._validate("tag", [], [], adjacent="none"),
            [
                fixture.error(
                    reports.codes.TAG_CANNOT_UPDATE_TAG_NO_IDS_SPECIFIED,
                ),
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_SPECIFY_ADJACENT_ID_WITHOUT_IDS_TO_ADD,
                    adjacent_id="none",
                ),
                fixture.error(
                    reports.codes.TAG_ADJACENT_REFERENCE_ID_NOT_IN_THE_TAG,
                    adjacent_id="none",
                    tag_id="tag",
                ),
            ],
        )

    def test_existing_tag_and_no_ids_with_existing_adjacent(self):
        assert_report_item_list_equal(
            self._validate("tag", [], [], adjacent="e1"),
            [
                fixture.error(
                    reports.codes.TAG_CANNOT_UPDATE_TAG_NO_IDS_SPECIFIED,
                ),
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_SPECIFY_ADJACENT_ID_WITHOUT_IDS_TO_ADD,
                    adjacent_id="e1",
                ),
            ],
        )

    def test_existing_tag_and_no_add_ids_with_existing_adjacent(self):
        assert_report_item_list_equal(
            self._validate("tag", [], ["e2"], adjacent="e1"),
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_SPECIFY_ADJACENT_ID_WITHOUT_IDS_TO_ADD,
                    adjacent_id="e1",
                ),
            ],
        )

    def test_add_ids_errors(self):
        assert_report_item_list_equal(
            self._validate(
                "tag",
                [
                    "new1",
                    "new1",
                    "none",
                    "none",
                    "other_id",
                    "other_id",
                    "e1",
                    "e1",
                    "status_id",
                    "status_id",
                    "e2",
                    "e2",
                ],
                [],
            ),
            [
                fixture.error(
                    reports.codes.TAG_ADD_REMOVE_IDS_DUPLICATION,
                    duplicate_ids_list=[
                        "e1",
                        "e2",
                        "new1",
                        "none",
                        "other_id",
                        "status_id",
                    ],
                    add_or_not_remove=True,
                ),
                fixture.report_not_found(
                    "none",
                    context_type="resources",
                    expected_types=self.OBJ_REF_EXPECTED_TYPES,
                ),
                fixture.report_unexpected_element(
                    "other_id", "meta_attributes", self.OBJ_REF_EXPECTED_TYPES
                ),
                fixture.report_not_found(
                    "status_id",
                    context_type="resources",
                    expected_types=self.OBJ_REF_EXPECTED_TYPES,
                ),
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_ADD_REFERENCE_IDS_ALREADY_IN_THE_TAG,
                    idref_list=["e1", "e2"],
                    tag_id="tag",
                ),
            ],
        )

    def test_non_existing_adjacent_in_add_ids(self):
        assert_report_item_list_equal(
            self._validate("tag", ["new1", "e1"], [], adjacent="e1"),
            [
                fixture.error(
                    reports.codes.TAG_CANNOT_PUT_ID_NEXT_TO_ITSELF,
                    adjacent_id="e1",
                ),
            ],
        )

    def test_existing_adjacent_in_add_ids(self):
        assert_report_item_list_equal(
            self._validate("tag", ["new1", "e1"], [], adjacent="new1"),
            [
                fixture.error(
                    reports.codes.TAG_ADJACENT_REFERENCE_ID_NOT_IN_THE_TAG,
                    adjacent_id="new1",
                    tag_id="tag",
                ),
            ],
        )

    def test_adjacent_in_remove_ids(self):
        assert_report_item_list_equal(
            self._validate("tag", ["e2"], ["e1"], adjacent="e1"),
            [
                fixture.error(
                    reports.codes.TAG_CANNOT_REMOVE_ADJACENT_ID,
                    adjacent_id="e1",
                ),
            ],
        )

    def test_remove_ids_errors(self):
        assert_report_item_list_equal(
            self._validate(
                "tag",
                [],
                [
                    "e1",
                    "e1",
                    "e2",
                    "e2",
                    "none",
                    "none",
                    "none1",
                    "none1",
                ],
            ),
            [
                fixture.error(
                    reports.codes.TAG_ADD_REMOVE_IDS_DUPLICATION,
                    duplicate_ids_list=["e1", "e2", "none", "none1"],
                    add_or_not_remove=False,
                ),
                fixture.error(
                    reports.codes.TAG_IDS_NOT_IN_THE_TAG,
                    tag_id="tag",
                    id_list=["none", "none1"],
                ),
            ],
        )

    def test_remove_all_without_add(self):
        assert_report_item_list_equal(
            self._validate("tag", [], ["e1", "e2", "e3"]),
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_REMOVE_REFERENCES_WITHOUT_REMOVING_TAG,
                    tag_id="tag",
                ),
            ],
        )

    def test_remove_all_and_add_new_one(self):
        assert_report_item_list_equal(
            self._validate("tag", ["new1"], ["e1", "e2", "e3"]),
            [],
        )

    def test_remove_all_and_add_missing_id(self):
        assert_report_item_list_equal(
            self._validate("tag", ["none"], ["e1", "e2", "e3"]),
            [
                fixture.report_not_found(
                    "none",
                    context_type="resources",
                    expected_types=self.OBJ_REF_EXPECTED_TYPES,
                ),
            ],
        )

    def test_add_and_remove_the_same_ids(self):
        assert_report_item_list_equal(
            self._validate("tag", ["new1"], ["new1"]),
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_ADD_AND_REMOVE_IDS_AT_THE_SAME_TIME,
                    idref_list=["new1"],
                ),
                fixture.error(
                    reports.codes.TAG_IDS_NOT_IN_THE_TAG,
                    tag_id="tag",
                    id_list=["new1"],
                ),
            ],
        )

    def test_move_and_remove_the_same_ids(self):
        assert_report_item_list_equal(
            self._validate("tag", ["e1", "e2"], ["e2", "e1"], adjacent="e3"),
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_ADD_AND_REMOVE_IDS_AT_THE_SAME_TIME,
                    idref_list=["e1", "e2"],
                ),
            ],
        )


class ExpandTag(TestCase):
    def setUp(self):
        self.cib = etree.fromstring(
            """
            <cib>
                <configuration>
                    <resources>
                        <primitive id="R1"/>
                        <clone id="C">
                            <primitive id="RC"/>
                        </clone>
                        <primitive id="R2"/>
                    </resources>
                    <tags>
                        <tag id="tag1">
                            <obj_ref id="R1"/>
                            <obj_ref id="C"/>
                        </tag>
                        <tag id="tag2">
                            <obj_ref id="R2"/>
                        </tag>
                    </tags>
                </configuration>
            </cib>
        """
        )

    @staticmethod
    def get_ids(elements):
        return [el.get("id") for el in elements]

    def test_no_tag(self):
        el_list = lib.expand_tag(self.cib.xpath(".//primitive[@id='R1']")[0])
        self.assertEqual(self.get_ids(el_list), ["R1"])

    def test_expand_all_types(self):
        el_list = lib.expand_tag(self.cib.xpath(".//tag[@id='tag1']")[0])
        self.assertEqual(self.get_ids(el_list), ["R1", "C"])

    def test_expand_selected_types(self):
        el_list = lib.expand_tag(
            self.cib.xpath(".//tag[@id='tag1']")[0], ["clone"]
        )
        self.assertEqual(self.get_ids(el_list), ["C"])



from unittest import TestCase

from pcs.cli.common import tools


class TimeoutToSecondsLegacyTest(TestCase):
    def test_valid(self):
        self.assertEqual(10, tools.timeout_to_seconds_legacy(10))
        self.assertEqual(10, tools.timeout_to_seconds_legacy("10"))
        self.assertEqual(10, tools.timeout_to_seconds_legacy("10s"))
        self.assertEqual(10, tools.timeout_to_seconds_legacy("10sec"))
        self.assertEqual(600, tools.timeout_to_seconds_legacy("10m"))
        self.assertEqual(600, tools.timeout_to_seconds_legacy("10min"))
        self.assertEqual(36000, tools.timeout_to_seconds_legacy("10h"))
        self.assertEqual(36000, tools.timeout_to_seconds_legacy("10hr"))

    def test_invalid(self):
        self.assertEqual(-10, tools.timeout_to_seconds_legacy(-10))
        self.assertEqual("1a1s", tools.timeout_to_seconds_legacy("1a1s"))
        self.assertEqual("10mm", tools.timeout_to_seconds_legacy("10mm"))
        self.assertEqual("10mim", tools.timeout_to_seconds_legacy("10mim"))
        self.assertEqual("aaa", tools.timeout_to_seconds_legacy("aaa"))
        self.assertEqual("", tools.timeout_to_seconds_legacy(""))



import base64
import json

from pcs.common import reports
from pcs.common.node_communicator import RequestData
from pcs.common.reports.item import ReportItem
from pcs.lib.communication.tools import (
    AllAtOnceStrategyMixin,
    AllSameDataMixin,
    RunRemotelyBase,
    SimpleResponseProcessingMixin,
    SkipOfflineMixin,
)


class BoothSendConfig(
    SimpleResponseProcessingMixin,
    SkipOfflineMixin,
    AllSameDataMixin,
    AllAtOnceStrategyMixin,
    RunRemotelyBase,
):
    # TODO adapt to new file transfer framework once it is written
    def __init__(
        self,
        report_processor,
        booth_name,
        config_data,
        authfile=None,
        authfile_data=None,
        skip_offline_targets=False,
    ):
        super().__init__(report_processor)
        self._set_skip_offline(skip_offline_targets)
        self._booth_name = booth_name
        self._config_data = config_data.decode("utf-8")
        self._authfile = authfile
        self._authfile_data = authfile_data

    def _get_request_data(self):
        data = {
            "config": {
                "name": "{0}.conf".format(self._booth_name),
                "data": self._config_data,
            }
        }
        if self._authfile is not None and self._authfile_data is not None:
            data["authfile"] = {
                "name": self._authfile,
                "data": base64.b64encode(self._authfile_data).decode("utf-8"),
            }
        return RequestData(
            "remote/booth_set_config", [("data_json", json.dumps(data))]
        )

    def _get_success_report(self, node_label):
        return ReportItem.info(
            reports.messages.BoothConfigAcceptedByNode(
                node=node_label,
                name_list=[self._booth_name],
            )
        )

    def before(self):
        self._report(
            ReportItem.info(reports.messages.BoothConfigDistributionStarted())
        )


class ProcessJsonDataMixin:
    __data = None

    @property
    def _data(self):
        if self.__data is None:
            self.__data = []
        return self.__data

    def _process_response(self, response):
        report = self._get_response_report(response)
        if report is not None:
            self._report(report)
            return
        target = response.request.target
        try:
            self._data.append((target, json.loads(response.data)))
        except ValueError:
            self._report(
                ReportItem.error(
                    reports.messages.InvalidResponseFormat(target.label)
                )
            )

    def on_complete(self):
        return self._data


class BoothGetConfig(
    ProcessJsonDataMixin,
    AllSameDataMixin,
    AllAtOnceStrategyMixin,
    RunRemotelyBase,
):
    def __init__(self, report_processor, booth_name):
        super().__init__(report_processor)
        self._booth_name = booth_name

    def _get_request_data(self):
        return RequestData(
            "remote/booth_get_config", [("name", self._booth_name)]
        )


class BoothSaveFiles(
    ProcessJsonDataMixin,
    AllSameDataMixin,
    AllAtOnceStrategyMixin,
    RunRemotelyBase,
):
    def __init__(self, report_processor, file_list, rewrite_existing=True):
        super().__init__(report_processor)
        self._file_list = file_list
        self._rewrite_existing = rewrite_existing
        self._output_data = []

    def _get_request_data(self):
        data = [("data_json", json.dumps(self._file_list))]
        if self._rewrite_existing:
            data.append(("rewrite_existing", "1"))
        return RequestData("remote/booth_save_files", data)

    def _process_response(self, response):
        report = self._get_response_report(response)
        if report is not None:
            self._report(report)
            return
        target = response.request.target
        try:
            parsed_data = json.loads(response.data)
            self._report(
                ReportItem.info(
                    reports.messages.BoothConfigAcceptedByNode(
                        node=target.label,
                        name_list=sorted(parsed_data["saved"]),
                    )
                )
            )
            for filename in list(parsed_data["existing"]):
                self._report(
                    ReportItem(
                        severity=reports.item.get_severity(
                            reports.codes.FORCE,
                            self._rewrite_existing,
                        ),
                        message=reports.messages.FileAlreadyExists(
                            # TODO specify file type; this will be overhauled
                            # to a generic file transport framework anyway
                            "",
                            filename,
                            node=target.label,
                        ),
                    )
                )
            for file, reason in dict(parsed_data["failed"]).items():
                self._report(
                    ReportItem.error(
                        reports.messages.BoothConfigDistributionNodeError(
                            target.label,
                            reason,
                            file,
                        )
                    )
                )
        except (KeyError, TypeError, ValueError):
            self._report(
                ReportItem.error(
                    reports.messages.InvalidResponseFormat(target.label)
                )
            )
